/*
 * com/ngon/mpeg/MPEGDecoder.java
 * Victor Liu > mailto:victor@n-gon.com
 *
 * This code is based on GPL'd code in MPEG_Play by Joerg Anders
 * which may be found at:
 *
 *     http://rnvs.informatik.tu-chemnitz.de/~jan/MPEG/MPEG_Play.html
 *
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version. 
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details. 
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

package com.ngon.mpeg;

import java.awt.Dimension;
import java.io.InputStream;
import java.io.IOException;
import java.lang.reflect.Constructor;

public class MPEGDecoder extends Thread {

    final static int PACK_START_CODE          = 0x000001ba;
    final static int SYSTEM_HEADER_START_CODE = 0x000001bb;
    final static int PACKET_START_CODE_PREFIX = 0x00000001;
    final static int ISO_11172_END_CODE       = 0x000001b9;

    final static int RESERVED_STREAM_ID  = 0xbc;
    final static int PRIVATE_STREAM_1_ID = 0xbd;
    final static int PADDING_STREAM_ID   = 0xbe;
    final static int PRIVATE_STREAM_2_ID = 0xbf;
    
    protected MPEGStream mpeg = null;
    protected volatile MPEGVideoDecoder vid = null;
    protected Semaphore videoSemaphore = new Semaphore();
    protected boolean hasVideoBegun = false;
    
    public MPEGDecoder(InputStream in) {
        this.mpeg = new MPEGStream(in, videoSemaphore);
        this.vid = new MPEGVideoDecoder(mpeg);
    }
    
    public MPEGDecoder(InputStream in, Class videoDecoderClass) {
        this.mpeg = new MPEGStream(in, videoSemaphore);
        try {
            Class[] types = new Class[1];
            types[0] = Class.forName("mpeg.MPEGStream");
            Constructor constructor = videoDecoderClass.getConstructor(types);
            Object[] args = new Object[1];
            args[0] = this.mpeg;
            this.vid = (MPEGVideoDecoder) constructor.newInstance(args);
        } catch (Exception e) {
            System.err.println(e);
        }
    }
    
    public MPEGFrame nextFrame() {
        if (vid != null) {
            return vid.nextFrame();
        } else {
            return null;
        }
    }
    
    public Dimension getVideoDimensions() {
        if (vid != null) {
            return new Dimension(vid.frameWidth, vid.frameHeight);
        } else {
            return null;
        }
    }
    
    public MPEGVideoDecoder getVideoDecoder() {
        return vid;
    }
    
    public void start() {
        if (vid == null) {
            System.err.println("No video decoder");
            return;
        }
        setPriority(Thread.MIN_PRIORITY);
        setDaemon(true);
        super.start();
    }
    
    public void run() {
        Thread currentThread = Thread.currentThread();
        try {
            mpeg.findNextStartCode();
            if (this == currentThread &&
                mpeg.compareNextBits(32, PACK_START_CODE)) {
                do {
                    eatPackHeader();
                    while (! mpeg.isEndOfFile() && 
                           ! mpeg.compareNextBits(32, PACK_START_CODE) &&
                           ! mpeg.compareNextBits(32, ISO_11172_END_CODE)) {
                        parsePacket();
                    }
                } while (this == currentThread &&
                         ! mpeg.isEndOfFile() && 
                         mpeg.compareNextBits(32, PACK_START_CODE));
            } else {
                // Many MPEG files only contain video (layer II) data.
                // In this case the video parser is called.
                vid.start();
                return;
            }
            
            if (this != currentThread) {
                return;
            }
            if (mpeg.compareNextBits(32, ISO_11172_END_CODE)) {
                System.out.println("ISO_11172_END_CODE found.");
            } else {
                int word = mpeg.getBits(32);
                System.out.println("ISO_11172_END_CODE missing (0x" + 
                                   Integer.toHexString(word) + ").");
            }
        } catch (Exception e) {
            System.err.println("MPEGDecoder: " + e);
        }
    }
    
    public void killVideoDecoder() {
        if (vid != null) {
            vid.disposeFrames();
            vid = null;
        }
    }
    
    private void eatPackHeader() throws IOException {
        mpeg.getBits(32);
        if (mpeg.getBits(4) != 0x2) {
            System.err.println("Sychronization error in PACK header.");
            return;
        }
        mpeg.getBits(30);
        mpeg.getBits(30);
        if (mpeg.compareNextBits(32, SYSTEM_HEADER_START_CODE)) {
            eatSystemHeader();
        }
    }

    private void eatSystemHeader() throws IOException {
        mpeg.getBits(30);
        mpeg.getBits(30);
        mpeg.getBits(30);
        mpeg.getBits(6);
        while (mpeg.compareNextBits(1, 1)) {
            mpeg.getBits(24);
        }
    }

    private void parsePacket() throws IOException {
        int streamId;
        int packetLength;
        boolean skipNonVideoTracks = true;
        
        if (mpeg.getBits(24) != 0x1) {
            System.err.println("Synchronization error in packet!");
            return;
        }
        streamId = mpeg.getBits(8);
        packetLength = mpeg.getBits(16);

        if (streamId != PRIVATE_STREAM_2_ID) {
            while (mpeg.compareNextBits(8, 0xff)) {
                mpeg.getBits(8);
                packetLength -= 1;
            }
            if (mpeg.compareNextBits(2, 0x1)) {
                mpeg.getBits(16);
                packetLength -= 2;
            }
            if (mpeg.compareNextBits(4, 0x2)) {
                mpeg.getBits(20);
                mpeg.getBits(20);
                packetLength -= 5;
            } else if (mpeg.compareNextBits(4, 0x3)) {
                mpeg.getBits(20);
                mpeg.getBits(20);
                mpeg.getBits(20);
                mpeg.getBits(20);
                packetLength -= 10;
            } else if (mpeg.getBits(8) != 0x0f) {
                System.err.println("Synchronization error in packet!");
            } else {
                packetLength--;
            }
        }
        if ((0xe0 & streamId) == 0xc0) { 
            // MPEG audio stream -> skip
        } else {
            switch (0xf0 & streamId) { 
            // MPEG video stream -> analyse
            case 0xe0:
                skipNonVideoTracks = false;
                
                // Subtract the bytes already in bit shift register.
                packetLength -= mpeg.getNumReadBytes();

                // Inform the input filter about the length of 
                // the video track.
                mpeg.setNumRemainingBytes(packetLength);

                if (! hasVideoBegun) {
                    hasVideoBegun = true;
                    vid.start();
                }
                // Pass control to video decoder thread and
                // wait until the video part is analysed.
                videoSemaphore.toggle1();
                break;
            case 0xf0: // RESERVED STREAM ID
                break;
            default: 
                switch (streamId) {
                case RESERVED_STREAM_ID: 
                    break;
                case PRIVATE_STREAM_1_ID: 
                    break;
                case PADDING_STREAM_ID: 
                    break;
                case PRIVATE_STREAM_2_ID: 
                    break;
                default: 
                    System.err.println("Unknown stream ID: 0x" + 
                                       Integer.toHexString(streamId));
                    return;
                }
            }
        }
        
        if (skipNonVideoTracks) {
            mpeg.skip(packetLength);
        }
    }
}
