/*
 * com/ngon/tap/Clip.java
 * Victor Liu > mailto:victor@n-gon.com
 *
 * Copyright (C) 2004 Victor Liu
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version. 
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details. 
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

package com.ngon.tap;

import com.ngon.app.App;
import com.ngon.mpeg.MPEGFrame;
import com.ngon.mpeg.MPEGDecoder;
import com.ngon.mpeg.MPEGVideoDecoder;
import java.awt.Dimension;
import java.io.InputStream;

/**
 * class Clip
 */
class Clip {
    
    protected int nStreams = 1;
    protected int nRetries = 10;
    
    protected String clipsPath;
    protected String clipName;
    protected App app;
    
    protected int clipID;
    protected int clipLength;
    protected int frameNr;
    
    protected InputStream[] mpegs;
    protected MPEGDecoder[] decs;
    protected MPEGFrame[] frames;
    protected int active = 0;
    
    protected Dimension frameDimensions;
    protected Dimension blockDimensions;
    
    Clip(String clipsPath, String clipName, App app) {
        this.clipsPath = clipsPath;
        this.clipName = clipName;
        this.app = app;
    }
    
    public void init() {
        mpegs = loadMPEGs(clipsPath, clipName, frameNr);
        if (mpegs == null) {
            return;
        }
        decs = createDecs(mpegs);
        frames = new MPEGFrame[nStreams];
    }
    
    public void start() {
        if (decs == null) {
            init();
            if (decs == null) {
                return;
            }
        }
        for (int i = 0; i < nStreams; i++) {
            if (decs[i] != null) {
                decs[i].start();
            }
        }
        if (frameNr > 0) {
            skipFrames(frameNr);
        }
    }
    
    public void skipFrames(int n) {
        for (int i = 0; i < n; i++) {
            if (! advanceDecs(30)) {
                System.err.println("Couldn't advanceDecs in skipFrames for i=" +
                    i);
            }
        }
    }
    
    public boolean nextFrame() {
        boolean success = advanceDecs(nRetries);
        if (success) {
            incrementFrameNr();
        }
        return success;
    }
    
    public void recordFrameStats(int flags) {
        if (decs == null) {
            init();
            if (decs == null) {
                return;
            }
        }
        for (int i = 0; i < nStreams; i++) {
            decs[i].getVideoDecoder().recordFrameStats(flags);
        }
    }
    
    public String[] getMPEGFilenames(String clipsPath, String clipName, 
                                     int frameNr) {
        String[] mpegFilenames = new String[nStreams];
        mpegFilenames[0] = new String(
            clipsPath + clipName + "/" + clipName + ".+0.mpg");
        return mpegFilenames;
    }
    
    public MPEGFrame getFrame() {
        return frames[active];
    }
    
    public void setClipID(int clipID) {
        this.clipID = clipID;
    }
    
    public int getClipID() {
        return this.clipID;
    }
    
    public void setClipLength(int clipLength) {
        this.clipLength = clipLength;
    }
    
    public int getClipLength() {
        return this.clipLength;
    }
    
    public void setFrameNr(int frameNr) {
        this.frameNr = frameNr;
    }
    
    public int getFrameNr() {
        return this.frameNr;
    }
    
    public void setNrRetries(int n) {
        this.nRetries = n;
    }
    
    public boolean isLastFrame() {
        return (this.frameNr >= this.clipLength-1);
    }
    
    public Dimension getFrameDimensions() {
        if (frameDimensions == null) {
            frameDimensions = decs[active].getVideoDimensions();
        }
        return frameDimensions;
    }
    
    public Dimension getBlockDimensions() {
        if (blockDimensions == null) {
            Dimension d = getFrameDimensions();
            blockDimensions = 
                new Dimension(
                    MPEGVideoDecoder.getMacroblockDimension(d.width), 
                    MPEGVideoDecoder.getMacroblockDimension(d.height));
        }
        return blockDimensions;
    }
    
    public int getBlockSize() {
        return MPEGVideoDecoder.getMacroblockSize();
    }
    
    public void kill() {
        for (int i = 0; i < nStreams; i++) {
            if (mpegs != null && mpegs[i] != null) {
                try {
                    mpegs[i].close();
                    mpegs[i] = null;
                } catch (Exception e) {
                    System.err.println("Exception closing mpeg : " + e);
                }
            }
            if (decs != null && decs[i] != null) {
                decs[i].killVideoDecoder();
                decs[i] = null;
            }
            if (frames != null && frames[i] != null) {
                frames[i] = null;
            }
        }
        mpegs = null;
        decs = null;
        frames = null;
        if (Debug.DEBUG || Debug.DEBUG_APPLET) {
            System.err.println("Killing clip " + clipID + "." + frameNr);
        }
    }
    
    protected InputStream[] loadMPEGs(String clipsPath, String clipName, 
                                      int frameNr) {
        String[] mpegFilenames = getMPEGFilenames(clipsPath, clipName, frameNr);
        InputStream[] mpegs = new InputStream[nStreams];
        for (int i = 0; i < nStreams; i++) {
            mpegs[i] = app.getDataStream(mpegFilenames[i]);
            if (mpegs[i] == null) {
                System.err.println("MPEG stream is null! " + i);
                return null;
            }
        }
        return mpegs;
    }
    
    protected MPEGDecoder[] createDecs(InputStream[] mpegs) {
        MPEGDecoder[] decs = new MPEGDecoder[nStreams];
        for (int i = 0; i < nStreams; i++) {
            decs[i] = new MPEGDecoder(mpegs[i]);
        }
        return decs;
    }
    
    protected boolean advanceDecs(int nTries) {
        if (decs == null) {
            System.err.println("Clip.advanceDecs : decs is null");
            return false;
        }
        for (int i = 0; i < nStreams; i++) {
            if (decs[i] == null) {
                System.err.println("Clip.advanceDecs : decs[" + i + 
                    "] is null");
                return false;
            }
            frames[i] = null;
            int cnt = 0;
            do {
                if ((frames[i] = decs[i].nextFrame()) != null) {
                    break;
                }
                if (++cnt > nTries) {
                    return false;
                }
                try {
                    Thread.sleep(50);
                    if (Debug.DEBUG) {
                        app.println("advanceDecs for " + clipID + "." +
                            frameNr + " : " + cnt);
                    }
                    if (Debug.DEBUG_APPLET) {
                        System.err.println("advanceDecs for " + clipID + "." +
                            frameNr + " : " + cnt);
                    }
                } catch (InterruptedException e) {}
                if (decs == null || decs[i] == null) {
                    return false;
                }
            } while (true);
        }
        return true;
    }
    
    protected void incrementFrameNr() {
        frameNr++;
    }
}

