/*
 * com/ngon/tap/TapStats.java
 * Victor Liu > mailto:victor@n-gon.com
 *
 * Copyright (C) 2004 Victor Liu
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version. 
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details. 
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

package com.ngon.tap;

import com.ngon.app.App;
import com.ngon.gfx.PixelSurface;
import com.ngon.gfx.ZoomPixelSurface;
import com.ngon.gfx.RGB;
import com.ngon.mpeg.MPEGDecoder;
import com.ngon.mpeg.MPEGFrame;
import com.ngon.mpeg.MPEGFrameStats;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.Point;
import java.sql.Connection;
import java.sql.DriverManager;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.Enumeration;
import java.util.Vector;

/**
 * class TapStats
 */
final public class TapStats extends TapApp {
    
    private String clipPath;
    private boolean doWait;
    private boolean doMV;
    private boolean doMVStore;
    private boolean doCG;
    
    private String clipName;
    private Clip2 clip;
    
    private PreparedStatement insertMVClusterStmt;
    private PreparedStatement insertMVBlockStmt;
    
    static public void main(String[] args) {
        if (args.length < 1) {
            printUsage();
            System.exit(0);
        }
        TapApp.main("TapStats", new TapStats(), args);
    }
    
    static public void printUsage() {
        System.out.println(
            "Usage: java -jar stats.jar <clipPath> [options]");
        System.out.println("Available options are");
        System.out.println("  -wait       Wait between frames");
        System.out.println("  -mv         Show motion vector clusters");
        System.out.println("  -cg         Show average positions");
        System.out.println("  -mv-store   Store motion vector clusters in db");
    }
    
    public void start() {
        clip.start();
        requestFocus();
        super.start();
    }

    public void main(int width, int height) {
        PixelSurface pixels = new ZoomPixelSurface(width, height, zoom);
        
        Thread currentThread = Thread.currentThread();
        while (_thread == currentThread) {
            
            if (doWait) {
                synchronized(this) {
                    waiting = true;
                }
            }
            
            if (! clip.nextFrame()) {
                if (doMVStore) {
                    // Record clip length.
                    storeClipLength();
                }
                // Exit.
                kill();
            }
            
            paintFrame(clip, pixels);
            
            int frameNr = clip.getFrameNr();
            if (frameNr < 2) {
                continue;
            }
            
            // Paint frame #.
            pixels.paintNumber(clip.getFrameNr()-1, 4, 2, 469, RGB.GRY);
                
            // Paint stats.
            MPEGFrame frame = clip.getFrame();
            MPEGFrameStats stats = frame.getStats();
            if (stats != null &&
                stats.isRecordingStat(MPEGFrameStats.MOTION_VECTORS)) {
                paintMV(clip, pixels);
                if (doMV || doCG) {
                    paintClusterStats(clip, pixels);
                }
            }
            
            update(pixels.getPixels());
            
            while (waiting) {
                try {
                     Thread.sleep(SLEEP);
                } catch (InterruptedException e) {}
            }
        }
    }
    
    public void stop() {
        if (clip != null) {
            clip.kill();
        }
        super.stop();
    }
    
    protected void getApplicationParameters(String[] args) {
        clipPath = args[0];
        for (int i = 1; i < args.length; i++) {
            if (args[i].equalsIgnoreCase("-wait")) {
                doWait = true;
            } else if (args[i].equalsIgnoreCase("-mv")) {
                doMV = true;
            } else if (args[i].equalsIgnoreCase("-cg")) {
                doCG = true;
            } else if (args[i].equalsIgnoreCase("-mv-store")) {
                doMV = true;
                doCG = true;
                doMVStore = true;
            }
        }
    }
    
    protected void getAppletParameters() {
        String s;
        if ((s = getParameter("clipPath")) != null) {
            clipPath = s;
        }
    }
    
    protected void initCommon() {
        setLayout(null);
        setBackground(Color.black);
        setVisible(true);
        initClip();
        initDBC();
    }
    
    private void initClip() {
        while (clipPath.charAt(clipPath.length()-1) == '/') {
            clipPath = clipPath.substring(0, clipPath.length()-1);
        }
        int p = clipPath.lastIndexOf("/");
        String clipsPath = clipPath.substring(0, p+1);
        clipName = clipPath.substring(p+1);
        clip = new Clip2(clipsPath, clipName, (App) this);
        clip.recordFrameStats(
            MPEGFrameStats.FRAME_TYPE | MPEGFrameStats.MOTION_VECTORS);
    }
    
    private void initDBC() {
        try {
            Class.forName("com.mysql.jdbc.Driver");
            dbc = DriverManager.getConnection(
                "jdbc:mysql://localhost/c3cuts?user=victor");
            
            if (doMVStore) {
                // Insert into clip table.
                Statement stmt = dbc.createStatement();
                stmt.executeUpdate(
                    "INSERT INTO clip (name) VALUES ('" + clipName + "')");
                
                // Get just-inserted clip id.
                ResultSet rs = stmt.executeQuery("SELECT LAST_INSERT_ID()");
                rs.next();
                clip.clipID = rs.getInt(1);
                
                // Create prepared INSERT statements for later.
                insertMVClusterStmt = 
                    dbc.prepareStatement(
                        "INSERT INTO mvcluster (" +
                        "n_blocks,width,height,cg_x,cg_y,avg_mv_x,avg_mv_y," +
                        "avg_div,avg_curl,avg_rgb,frame_nr,clip_id) " + 
                        "VALUES (?,?,?,?,?,?,?,?,?,?,?," + clip.clipID + ")");
                insertMVBlockStmt = 
                    dbc.prepareStatement(
                        "INSERT INTO mvblock (" +
                        "block_nr,mv_x,mv_y,mvcluster_id) VALUES (?,?,?,?)");
            }
        } catch (Exception e) {
            System.err.println(e);
        }
    }
    
    private void storeClipLength() {
        try {
            Statement stmt = dbc.createStatement();
            stmt.executeUpdate(
                "UPDATE clip SET len=" + clip.getFrameNr() + 
                " WHERE id=" + clip.clipID);
        } catch (Exception e) {
            System.err.println(e);
        }
    }
    
    /* ========================================================================
        MV OPS
       ========================================================================
     */
    private void paintMV(Clip2 clip, PixelSurface pixels) {
        MPEGFrame frame = clip.getFrame();
        MPEGFrameStats stats = frame.getStats();
        Dimension mbDim = stats.mbDimensions;
        Point[] mv = stats.motionVectors;
        int p = 0;
        for (int j = 0; j < mbDim.height; j++) {
            for (int i = 0; i < mbDim.width; i++) {
                if (mv[p] != null) {
                    paintMV(origin.x + (i * 16 + 7) * zoom, 
                            origin.y + (j * 16 + 7) * zoom, 
                            mv[p], 255, pixels);
                }
                p++;
            }
        }
    }
    
    private void paintClusterStats(Clip2 clip, PixelSurface pixels) {
        MPEGFrame frame = clip.getFrame();
        MPEGFrameStats stats = frame.getStats();
        Dimension mbDim = stats.mbDimensions;
        Point[] mv = stats.motionVectors;
        Vector mvClusters = ClusStats.findClusters(frame);
        int n = mvClusters.size();
        int frameNr = clip.getFrameNr();
        int count = 0;
        for (int i = 0; i < n; i++) {
            ClusStats c = (ClusStats) mvClusters.elementAt(i);
            count++;
            boolean isSpurious = c.isSpurious();
            int rgb = isSpurious ? RGB.GRY 
                                 : CLUSTER_RGB[count % CLUSTER_RGB.length];
            int clusterSize = c.size();
            
            if (doMV) {
                for (int j = 0; j < clusterSize; j++) {
                    int idx = ((Integer) c.elementAt(j)).intValue();
                    int x = origin.x + ((idx % mbDim.width) * 16 + 0) * zoom;
                    int y = origin.y + ((idx / mbDim.width) * 16 + 0) * zoom;
                    paintBlockOutline(x, y, rgb, pixels);
                }
            }
            
            Point cg = c.getCenterOfGravity();
            Point avgMV = c.getAverageMV();
            int avgDiv = c.getAverageDiv();
            int avgCurl = c.getAverageCurl();
            int avgRGB = c.getAverageRGB();
            
            if (doCG) {
                // Paint cluster average MV and CG.
                int x = cg.x * zoom;
                int y = cg.y * zoom;
                int dx = -((avgMV.x * zoom) >> 1);
                int dy = -((avgMV.y * zoom) >> 1);
                int crossRGB = isSpurious ? 0xCC6666 : RGB.RED;
                int mvRGB = isSpurious ? 0x999966 : RGB.YLW;
                int rgb1 = isSpurious ? 0x669966 : RGB.GRN;
                int rgb2 = isSpurious ? 0x996699 : RGB.PRP;
                paintAvg(x, y, dx, dy, rgb, crossRGB, mvRGB, pixels);
//                pixels.paintNumber(i, 2, x + 6, y + 2, digitRGB);
                pixels.paintNumber(-avgDiv, 2, x - 16, y - 9, rgb1);
                pixels.paintNumber(-avgCurl, 2, x + 5, y - 9, rgb2);
            }
            
            if (doMVStore && ! isSpurious) {
                Dimension clusterDim = c.getClusterDimensions();
                storeClusterStats(clusterSize, clusterDim, 
                                  cg, avgMV, avgDiv, avgCurl, avgRGB, frameNr, 
                                  c.elements(), mv);
            }
        }
    }
    
    private void storeClusterStats(int clusterSize, Dimension clusterDim,
                                   Point cg, Point avgMV, int avgDiv, 
                                   int avgCurl, int avgRGB, int frameNr, 
                                   Enumeration blocks, Point[] mv) {
        try {
            // Store aggregate data for cluster.
            insertMVClusterStmt.setInt(1, clusterSize);
            insertMVClusterStmt.setInt(2, clusterDim.width);
            insertMVClusterStmt.setInt(3, clusterDim.height);
            insertMVClusterStmt.setInt(4, cg.x);
            insertMVClusterStmt.setInt(5, cg.y);
            insertMVClusterStmt.setInt(6, -avgMV.x);
            insertMVClusterStmt.setInt(7, -avgMV.y);
            insertMVClusterStmt.setInt(8, -avgDiv);
            insertMVClusterStmt.setInt(9, -avgCurl);
            insertMVClusterStmt.setInt(10, avgRGB);
            insertMVClusterStmt.setInt(11, frameNr-1);
            insertMVClusterStmt.execute();
            
            // Store block data for cluster.
            Statement stmt = dbc.createStatement();
            ResultSet rs = stmt.executeQuery("SELECT LAST_INSERT_ID()");
            rs.next();
            int clusterID = rs.getInt(1);
            while (blocks.hasMoreElements()) {
                int blockNr = ((Integer) blocks.nextElement()).intValue();
                insertMVBlockStmt.setInt(1, blockNr);
                insertMVBlockStmt.setInt(2, -mv[blockNr].x);
                insertMVBlockStmt.setInt(3, -mv[blockNr].y);
                insertMVBlockStmt.setInt(4, clusterID);
                insertMVBlockStmt.execute();
            }
        } catch (SQLException e) {
            System.err.println(e);
        }
    }
}
